/***************************************************************************
                          plugin_katesymbolviewer.cpp  -  description
                             -------------------
    begin                : Apr 2 2003
    author               : 2003 Massimo Callegari
    email                : massimo_callegari@libero.it
    
    Changes:
    Apr 07 2003 v 0.1 - First version.
    Apr 19 2003 v 0.2 - Added to CVS. Extract functions and structures
    Apr 22 2003 v 0.3 - Added macro extraction + several bugfixing
    Apr 23 2003 v 0.4 - "View Symbol" moved in Settings menu. "Refresh List" is no
                        longer in Kate menu. Moved into a popup menu activated by a
                        mouse right button click. + Bugfixing.
    Apr 24 2003 v 0.5 - Added three check buttons in popup menu to show/hide symbols
    May 04 2003 v 0.6 - Symbol List becomes a KListView object. Removed Tooltip class.
                        Added a QTimer that every 200ms checks:
                        * if the list width has changed
                        * if the document has changed
                        Added an entry in the popup menu to switch between List and Tree mode
                        Various bugfixing.
    Jun 19 2003 v.1.0 - Removed QTimer (polling is Evil(tm)... )
                        - Captured documentChanged() event to refresh symbol list
                        - Tooltips vanished into nowhere...sigh :(
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
 
#include "plugin_katesymbolviewer.h"
#include "plugin_katesymbolviewer.moc"

#include <kaction.h>
#include <klocale.h>
#include <kstandarddirs.h>

/* XPM */
static const char* const class_xpm[] = { 
"16 16 10 1",
" 	c None",
".	c #000000",
"+	c #A4E8FC",
"@	c #24D0FC",
"#	c #001CD0",
"$	c #0080E8",
"%	c #C0FFFF",
"&	c #00FFFF",
"*	c #008080",
"=	c #00C0C0",
"     ..         ",
"    .++..       ",
"   .+++@@.      ",
"  .@@@@@#...    ",
"  .$$@@##.%%..  ",
"  .$$$##.%%%&&. ",
"  .$$$#.&&&&&*. ",
"   ...#.==&&**. ",
"   .++..===***. ",
"  .+++@@.==**.  ",
" .@@@@@#..=*.   ",
" .$$@@##. ..    ",
" .$$$###.       ",
" .$$$##.        ",
"  ..$#.         ",
"    ..          "};

static const char* const struct_xpm[] = {
"16 16 14 1",
" 	c None",
".	c #000000",
"+	c #C0FFC0",
"@	c #00FF00",
"#	c #008000",
"$	c #00C000",
"%	c #C0FFFF",
"&	c #00FFFF",
"*	c #008080",
"=	c #00C0C0",
"-	c #FFFFC0",
";	c #FFFF00",
">	c #808000",
",	c #C0C000",
"     ..         ",
"    .++..       ",
"   .+++@@.      ",
"  .@@@@@#...    ",
"  .$$@@##.%%..  ",
"  .$$$##.%%%&&. ",
"  .$$$#.&&&&&*. ",
"   ...#.==&&**. ",
"   .--..===***. ",
"  .---;;.==**.  ",
" .;;;;;>..=*.   ",
" .,,;;>>. ..    ",
" .,,,>>>.       ",
" .,,,>>.        ",
"  ..,>.         ",
"    ..          "};

// Sarah's colors...
static const char* const macro_xpm[] = {
"16 16 14 1",
" 	c None",
".	c #000000",
"+	c #FF7FE5",
"@	c #FF00C7",
"#	c #7F0066",
"$	c #BC0096",
"%	c #C0FFFF",
"&	c #00FFFF",
"*	c #008080",
"=	c #00C0C0",
"-	c #D493FF",
";	c #A100FF",
">	c #470082",
",	c #6B00B7",
"     ..         ",
"    .++..       ",
"   .+++@@.      ",
"  .@@@@@#...    ",
"  .$$@@##.%%..  ",
"  .$$$##.%%%&&. ",
"  .$$$#.&&&&&*. ",
"   ...#.==&&**. ",
"   .--..===***. ",
"  .---;;.==**.  ",
" .;;;;;>..=*.   ",
" .,,;;>>. ..    ",
" .,,,>>>.       ",
" .,,,>>.        ",
"  ..,>.         ",
"    ..          "};

class PluginView : public KXMLGUIClient
{             
  friend class KatePluginSymbolViewer;

  public:
    Kate::MainWindow *win;
};

extern "C"
{
  void* init_katecppsymbolviewerplugin()
  {
    KGlobal::locale()->insertCatalogue("katesymbolviewer"); 
    return new KatePluginFactory;
  }
}

KatePluginFactory::KatePluginFactory()
{
  s_instance = new KInstance( "kate" );
}

KatePluginFactory::~KatePluginFactory()
{
  delete s_instance;
}

QObject* KatePluginFactory::createObject( QObject* parent, const char* name, const char*, const QStringList & )
{
  return new KatePluginSymbolViewer( parent, name );
}

KInstance* KatePluginFactory::s_instance = 0L;

KateSymbolList::KateSymbolList ()
{

}

KatePluginSymbolViewer::KatePluginSymbolViewer( QObject* parent, const char* name )
    : Kate::Plugin ( (Kate::Application*)parent, name )
{
 m_Active = false;
 popup = new QPopupMenu();
 popup->insertItem(i18n("Refresh List"), this, SLOT(slotRefreshSymbol()));
 popup->insertSeparator();
 m_macro = popup->insertItem(i18n("Show Macros"), this, SLOT(toggleShowMacros()));
 m_struct = popup->insertItem(i18n("Show Structures"), this, SLOT(toggleShowStructures()));
 m_func = popup->insertItem(i18n("Show Functions"), this, SLOT(toggleShowFunctions()));
 popup->insertSeparator();
 popup->insertItem(i18n("List/Tree Mode"), this, SLOT(slotChangeMode()));
 
 popup->setItemChecked(m_macro, true);
 popup->setItemChecked(m_struct, true);
 popup->setItemChecked(m_func, true);
 macro_on = true;
 struct_on = true;
 func_on = true;
 
}

void KatePluginSymbolViewer::toggleShowMacros(void)
{
 bool s = !popup->isItemChecked(m_macro);
 popup->setItemChecked(m_macro, s);
 macro_on = s;
 slotRefreshSymbol();
}

void KatePluginSymbolViewer::toggleShowStructures(void)
{
 bool s = !popup->isItemChecked(m_struct);
 popup->setItemChecked(m_struct, s);
 struct_on = s;
 slotRefreshSymbol();
}

void KatePluginSymbolViewer::toggleShowFunctions(void)
{
 bool s = !popup->isItemChecked(m_func);
 popup->setItemChecked(m_func, s);
 func_on = s;
 slotRefreshSymbol();
}

KatePluginSymbolViewer::~KatePluginSymbolViewer()
{
}

void KatePluginSymbolViewer::addView(Kate::MainWindow *win)
{
    PluginView *view = new PluginView ();
             
     (void) new KToggleAction ( i18n("View Symbols"), 0, this,
                      SLOT( slotInsertSymbol() ), view->actionCollection(),
                      "view_insert_symbolviewer" );
                      
    view->setInstance (new KInstance("kate"));
    view->setXMLFile("plugins/katecppsymbolviewer/ui.rc");
    win->guiFactory()->addClient (view);
    view->win = win; 
    
   m_views.append (view);
}   

void KatePluginSymbolViewer::removeView(Kate::MainWindow *win)
{
  for (uint z=0; z < m_views.count(); z++)
    if (m_views.at(z)->win == win)
    {
      PluginView *view = m_views.at(z);
      m_views.remove (view);
      win->guiFactory()->removeClient (view);
      delete view;
    }  
}

void KatePluginSymbolViewer::slotInsertSymbol()
{
 QPixmap cls( ( const char** ) class_xpm );
 Kate::ToolViewManager *kt = application()->activeMainWindow()->toolViewManager();
 Kate::DocumentManager *kd = application()->documentManager();
 
 if (m_Active == false)
     {
      symbols = new KateSymbolList();
      listMode = 0;
      
      connect(symbols, SIGNAL(executed(QListViewItem *)), this, SLOT(goToSymbol(QListViewItem *)));
      connect(symbols, SIGNAL(rightButtonClicked(QListViewItem *, const QPoint&, int)),
               SLOT(slotShowContextMenu(QListViewItem *, const QPoint&, int)));
      connect(kd, SIGNAL(documentChanged()), this, SLOT(slotDocChanged()));
      connect(symbols, SIGNAL(resizeEvent(QResizeEvent * e)), this, SLOT(slotViewChanged(QResizeEvent * e)));
      
      if (kt)
          kt->addToolViewWidget(KDockWidget::DockLeft, symbols, cls, i18n("Symbol List"));
          
      m_Active = true;
      //symbols->addColumn(i18n("Symbols"), symbols->parentWidget()->width());
      symbols->addColumn(i18n("Symbols"));
      symbols->setSorting(-1, FALSE);
      symbols->setRootIsDecorated(0);
      symbols->setTreeStepSize(10);
      symbols->setShowToolTips(TRUE);
     
      /* First Symbols parsing here...*/
      parseSymbols();
     }
  else
     {
      if (kt)
        kt->removeToolViewWidget(symbols);
      m_Active = false;
     }
}

void KatePluginSymbolViewer::slotRefreshSymbol()
{
 symbols->clear();
 parseSymbols();
}

void KatePluginSymbolViewer::slotChangeMode()
{
 listMode = !listMode;
 symbols->clear();
 parseSymbols();
}

void KatePluginSymbolViewer::slotDocChanged()
{
 //kdDebug(13000)<<"Document changed !!!!"<<endl;
 slotRefreshSymbol();
}

void KatePluginSymbolViewer::slotViewChanged(QResizeEvent * e)
{
 kdDebug(13000)<<"View changed !!!!"<<endl;
 symbols->setColumnWidth(0, symbols->parentWidget()->width());
}

void KatePluginSymbolViewer::slotShowContextMenu(QListViewItem *it, const QPoint &p, int col)
{
 popup->popup(p);
}

void KatePluginSymbolViewer::parseSymbols(void)
{
 QString cl; // Current Line
 QString stripped;
 uint i, j, tmpPos = 0; 
 int par = 0, graph = 0, retry = 0;
 char block = 0, comment = 0, macro = 0, macro_pos = 0;
 bool structure = false;
 QPixmap cls( ( const char** ) class_xpm );
 QPixmap sct( ( const char** ) struct_xpm );
 QPixmap mcr( ( const char** ) macro_xpm );
 QListViewItem *node = NULL;
 QListViewItem *mcrNode = NULL, *sctNode = NULL, *clsNode = NULL;
 QListViewItem *lastMcrNode = NULL, *lastSctNode = NULL, *lastClsNode = NULL;
 
 Kate::Document *kv = application()->activeMainWindow()->viewManager()->activeView()->getDoc();
 //kdDebug(13000)<<"Lines counted :"<<kv->numLines()<<endl;
 positions.resize(kv->numLines() + 3); // Maximum symbols number o.O
 positions.fill(0);
 if(listMode)
   {
    mcrNode = new QListViewItem(symbols, symbols->lastItem(), i18n("Macros"));
    sctNode = new QListViewItem(symbols, symbols->lastItem(), i18n("Structures"));
    clsNode = new QListViewItem(symbols, symbols->lastItem(), i18n("Functions"));
    lastMcrNode = mcrNode;
    lastSctNode = sctNode;
    lastClsNode = clsNode;
    symbols->setRootIsDecorated(1);
   }
 else symbols->setRootIsDecorated(0);
     
 for (i=0; i<kv->numLines(); i++)
   {
    cl = kv->textLine(i);
    cl = cl.stripWhiteSpace();
    if(cl.find("/*") >= 0 && graph == 0) comment = 1;
    if(cl.find("*/") >= 0 && graph == 0) comment = 0;
    if(cl.find("#") >= 0 && graph == 0 ) macro = 1;
    if (comment != 1)
      {
       /* *********************** MACRO PARSING *****************************/
       if(macro == 1)
         {
          //macro_pos = cl.find("#");
          for (j = 0; j < cl.length(); j++)
             {
              if(cl.at(j)=='/' && cl.at(j+1)=='/') { macro = 0; break; }
              if((uint)cl.find("define") == j)
                    {
                     macro = 2;
                     j += 6; // skip the word "define"
                    }
              if(macro == 2 && cl.at(j) != ' ') macro = 3;
              if(macro == 3) 
                { 
                 if (cl.at(j) >= 0x20) stripped += cl.at(j);
                 if (cl.at(j) == ' ' || j == cl.length() - 1) 
                         macro = 4;
                }
              //kdDebug(13000)<<"Macro -- Stripped : "<<stripped<<" macro = "<<macro<<endl;
             }
           // I didn't find a valid macro e.g. include
           if(j == cl.length() && macro == 1) macro = 0; 
           if(macro == 4)
             {
              //stripped.replace(0x9, " ");
              stripped = stripped.stripWhiteSpace();
              if (macro_on == true)
                 {
                  if (listMode) 
                    {
                     node = new QListViewItem(mcrNode, lastMcrNode, stripped);
                     lastMcrNode = node;
                    }
                  else node = new QListViewItem(symbols, symbols->lastItem(), stripped);
                  node->setPixmap(0, (const QPixmap &)mcr);
                  positions.at(symbols->itemIndex(node)) = i;
                 }
              macro = 0;
              macro_pos = 0;
              stripped = "";
              //kdDebug(13000)<<"Macro -- Inserted : "<<stripped<<" at row : "<<i<<endl;
              if (cl.at(cl.length() - 1) == '\\') macro = 5; // continue in rows below
              continue;
             }
          }
       if (macro == 5 && cl.at(cl.length() - 1) != '\\') { macro = 0; continue; }
       
       /* ******************************************************************** */
         
       if(cl.find("(") >= 0 && cl.at(0) != '#' && block == 0) 
          { structure = false; block = 1; }
       if((cl.find("typedef") >= 0 || cl.find("struct") >= 0) && 
          graph == 0 && block == 0)
         { structure = true; block = 2; stripped = ""; }
       //if(cl.find(";") >= 0 && graph == 0)
       //    block = 0;
       if(block > 0)
         {
          for (j = 0; j < cl.length(); j++)
            {
             if (cl.at(j) == '/' && (cl.at(j + 1) == '*')) comment = 2;
             if (cl.at(j) == '*' && (cl.at(j + 1) == '/')) {  comment = 0; j+=2; }
             if(cl.at(j)=='/' && cl.at(j+1)=='/') 
               { if(block == 1 && stripped == "") block = 0; break; }
             if (comment != 2)
               {
                if (block == 1 && graph == 0 ) 
                  {
                   if (cl.at(j) >= 0x20) stripped += cl.at(j);
                   if(cl.at(j) == '(') par++;
                   if(cl.at(j) == ')') 
                     {
                      par--;
                      if(par == 0)
                        {
                         stripped = stripped.stripWhiteSpace();
                         stripped.remove("static ");
                         //kdDebug(13000)<<"Function -- Inserted : "<<stripped<<" at row : "<<i<<endl;
                         block = 2;
                         tmpPos = i;
                        }
                     }
                  } // BLOCK 1
                if(block == 2 && graph == 0)
                  {
                   if(cl.at(j)=='/' && cl.at(j+1)=='/') break;
                   if(cl.at(j)==':' || cl.at(j)==',') { block = 1; continue; }
                   if(cl.at(j)==';')
                     {
                      stripped = "";
                      block = 0;
                      structure = false;
                      break;
                     } 
                   
                   if(cl.at(j)=='{' && structure == false && cl.find(";") < 0)
                     {
                      stripped.replace(0x9, " ");
                      if(func_on == true)
                        {
                         if (listMode) 
                           { 
                            node = new QListViewItem(clsNode, lastClsNode, stripped); 
                            lastClsNode = node; 
                           }
                         else node = new QListViewItem(symbols, symbols->lastItem(), stripped);
                         node->setPixmap(0, (const QPixmap &)cls);
                         positions.at(symbols->itemIndex(node)) = tmpPos;
                        }
                      stripped = "";
                      retry = 0;
                      block = 3;
                     }
                   if(cl.at(j)=='{' && structure == true)
                     { 
                      block = 3;
                      tmpPos = i;
                     } 
                   if(cl.at(j)=='(')
                     {
                      retry = 1;
                      block = 0;
                      j = 0; 
                      kdDebug(13000)<<"Restart from the beginnin of line..."<<endl;
                      stripped = "";
                      if (retry == 1) { break; } // Avoid an infinite loop :(
                      else retry = 1;
                     }
                   if(structure == true && cl.at(j) >= 0x20) stripped += cl.at(j);
                  } // BLOCK 2
                
                if (block == 3)
                  {
                   // A comment...there can be anything
                   if(cl.at(j)=='/' && cl.at(j+1)=='/') break;
                   if(cl.at(j)=='{') graph++;
                   if(cl.at(j)=='}') 
                     {
                      graph--;
                      if (graph == 0 && structure == false) block = 0;
                      if (graph == 0 && structure == true) block = 4;
                     }
                  } // BLOCK 3
                
                if (block == 4)
                  {
                   if(cl.at(j) == ';')
                     {
                      //stripped.replace(0x9, " ");
                      stripped.remove('{');
                      stripped.replace('}', " ");
                      if(struct_on == true)
                        {
                         if (listMode) 
                           {
                            node = new QListViewItem(sctNode, lastSctNode, stripped);
                            lastSctNode = node;
                           } 
                         else node = new QListViewItem(symbols, symbols->lastItem(), stripped);
                         node->setPixmap(0, (const QPixmap &)sct);
                         positions.at(symbols->itemIndex(node)) = tmpPos;
                        }
                      //kdDebug(13000)<<"Structure -- Inserted : "<<stripped<<" at row : "<<i<<endl;
                      stripped = "";
                      block = 0;
                      structure = false;
                      //break;
                      continue;
                     }
                   if (cl.at(j) >= 0x20) stripped += cl.at(j);
                  } // BLOCK 4
               } // comment != 2
            } // End of For cycle
         } // BLOCK > 0
      } // Comment != 1
   } // for kv->numlines

 positions.resize(symbols->itemIndex(node) + 1);
 //for (i= 0; i < (symbols->itemIndex(node) + 1); i++)
 //    kdDebug(13000)<<"Symbol row :"<<positions.at(i) <<endl;
}

void KatePluginSymbolViewer::goToSymbol(QListViewItem *it)
{
 Kate::View *kv = application()->activeMainWindow()->viewManager()->activeView();
 kdDebug(13000)<<"Slot Activated at pos: "<<symbols->itemIndex(it) <<endl;
 kv->gotoLineNumber(positions.at(symbols->itemIndex(it)) + 10);
 kv->setFocus();
 kv->gotoLineNumber(positions.at(symbols->itemIndex(it)));
}



