/***************************************************************************
                          plugin_katesymbolviewer.h  -  description
                             -------------------
    begin                : Apr 2 2003
    author               : 2003 Massimo Callegari
    email                : massimo_callegari@libero.it
 ***************************************************************************/
 
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
 
#ifndef _PLUGIN_KATE_SYMBOLVIEWER_H_
#define _PLUGIN_KATE_SYMBOLVIEWER_H_

#include <kate/application.h>
#include <kate/documentmanager.h>
#include <kate/document.h>
#include <kate/mainwindow.h>
#include <kate/plugin.h>
#include <kate/view.h>
#include <kate/viewmanager.h>
#include <kate/toolviewmanager.h>

#include <kdebug.h>
#include <qmemarray.h>
#include <qpopupmenu.h>
#include <qevent.h>
#include <klibloader.h>
#include <klocale.h>
#include <klistview.h>
#include <kiconloader.h>

class KatePluginFactory : public KLibFactory
{
  Q_OBJECT

  public:
    KatePluginFactory();
    virtual ~KatePluginFactory();

    virtual QObject* createObject( QObject* parent = 0, const char* pname = 0, const char* name = "QObject", const QStringList &args = QStringList() );

  private:
    static KInstance* s_instance;
};

class KatePluginSymbolViewer : public Kate::Plugin, Kate::PluginViewInterface
{
  Q_OBJECT

  public:
    KatePluginSymbolViewer( QObject* parent = 0, const char* name = 0 );
    virtual ~KatePluginSymbolViewer();
    
    void addView (Kate::MainWindow *win);
    void removeView (Kate::MainWindow *win);
    
  public slots:
    void slotInsertSymbol();
    void slotRefreshSymbol();
    void slotChangeMode();
    void slotDocChanged();
    void goToSymbol(QListViewItem *);
    void slotShowContextMenu(QListViewItem *, const QPoint&, int);
    void toggleShowMacros(void);
    void toggleShowStructures(void);
    void toggleShowFunctions(void);
  protected:
    void slotViewChanged(QResizeEvent * e);
  private:
    QPtrList<class PluginView> m_views; 
    QPopupMenu *popup;
    KListView *symbols;
    bool m_Active;
    int m_macro, m_struct, m_func;
    bool macro_on, struct_on, func_on;
    bool listMode;
    QMemArray<uint> positions;
    void parseSymbols(void);

};

class KateSymbolList : public KListView
{
  Q_OBJECT
  
  public:
    KateSymbolList ();
    ~KateSymbolList () { };
  
  public slots:
  
  protected:

};


#endif
